from .ecdf import calc_ecdf_dist, ecdf, ecdf_distance
from .hist import calc_hist_dist, hist_distance, histcdf
from .scaled_w1 import calc_sw1_dist, scaled_w1_distance
from .wasserstein import energy_distance, wasserstein_distance
